import os
from PIL import Image
import Augmentor
import matplotlib.pyplot as plt
import cupy as np
import time

import cupy as np
import scipy
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.ndimage import rotate



def show_img(img, ax):
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.imshow(img)


def plot_grid(imgs, nrows, ncols, figsize=(10, 10)):
    assert len(imgs) == nrows * ncols, f"Number of images should be {nrows}x{ncols}"
    _, axs = plt.subplots(nrows, ncols, figsize=figsize)
    axs = axs.flatten()
    for img, ax in zip(imgs, axs):
        show_img(img, ax)


#

base_dir = '/jjy/dataset/dog_cat'

train_dir = os.path.join(base_dir, 'training_set\\training_set')
test_dir = os.path.join(base_dir, 'test_set\\test_set')

# 훈련에 사용되는 고양이/개 이미지 경로
train_cats_dir = os.path.join(train_dir, 'cats')
train_dogs_dir = os.path.join(train_dir, 'dogs')

p = Augmentor.Pipeline(train_cats_dir)
p.random_distortion(probability=1, grid_width=4, grid_height=4, magnitude=8)
p.flip_left_right(probability=0.5)
p.zoom(probability=0.5, min_factor=1.1, max_factor=1.25)
p.rotate(probability=0.7, max_left_rotation=10, max_right_rotation=10)
p.crop_random(probability=0.5, percentage_area=0.9)
p.sample(10000)




# %%

from jjy.framework.cuda import xp as np


def _numerical_gradient_1d(f, x):
    h = 1e-4  # 0.0001
    grad = np.zeros_like(x)

    for idx in range(x.size):
        tmp_val = x[idx]
        x[idx] = float(tmp_val) + h
        fxh1 = f(x)  # f(x+h)

        x[idx] = tmp_val - h
        fxh2 = f(x)  # f(x-h)
        grad[idx] = (fxh1 - fxh2) / (2 * h)

        x[idx] = tmp_val  # 값 복원

    return grad


def numerical_gradient_2d(f, X):
    if X.ndim == 1:
        return _numerical_gradient_1d(f, X)
    else:
        grad = np.zeros_like(X)

        for idx, x in enumerate(X):
            grad[idx] = _numerical_gradient_1d(f, x)

        return grad


def numerical_gradient(f, x):
    h = 1e-4  # 0.0001
    grad = np.zeros_like(x)

    it = np.nditer(x, flags=['multi_index'], op_flags=['readwrite'])

    """
    * 2 x 4행렬의 경우

(0, 0) -> (0, 1) -> (0, 2) -> (0, 3) -> (1, 0) -> (1, 1) -> (1, 2) -> (1, 3) 순으로 접근
    """
    while not it.finished:
        idx = it.multi_index
        #         print(idx)

        tmp_val = x[idx]
        x[idx] = float(tmp_val) + h
        fxh1 = f(x)  # f(x+h)

        x[idx] = tmp_val - h
        fxh2 = f(x)  # f(x-h)
        grad[idx] = (fxh1 - fxh2) / (2 * h)

        x[idx] = tmp_val  # 값 복원
        it.iternext()

    return grad



# %%

import jjy.framework.cuda as cuda
import jjy.framework.initializer as Initializer
from jjy.framework.cuda import xp as np
from jjy.framework.functions import *
from jjy.framework.gradient import *






class LayerType:
    def __init__(self, **kwargs):


        self.trainable = True
        if "trainable" in kwargs:
            self.trainable = kwargs["trainable"]
        else:
            self.trainable = True



    def set_trainable(self, val):
        self.trainable = val

    def get_trainable(self):
        return self.trainable

class FreezePrev(LayerType):

    def __init__(self):
        pass


class Dense(LayerType):

    def __init__(self, hidden_size, **kwargs):
        super().__init__(**kwargs)
        self.hidden_size = hidden_size
        if "input_size" in kwargs:
            self.input_size = kwargs["input_size"]
        else:
            self.input_size = None
        if "activation" in kwargs:
            self.activation = kwargs["activation"]
        else:
            self.activation = None

        if "initializer" in kwargs:
            self.initializer = kwargs["initializer"]
        else:
            self.initializer = Initializer.Std(0.01)


class Relu(LayerType):
    def __init__(self):
        super().__init__()
        self.mask = None
        self.activation = True

    def forward(self, x):
        self.mask = (x <= 0)
        out = x.copy()
        out[self.mask] = 0

        return out

    def backward(self, dout):
        dout[self.mask] = 0
        dx = dout

        return dx


class Sigmoid(LayerType):
    def __init__(self):
        super().__init__()
        self.out = None
        self.activation = True

    def forward(self, x):
        out = sigmoid(x)
        self.out = out
        return out

    def backward(self, dout):
        dx = dout * (1.0 - self.out) * self.out

        return dx






class Affine(LayerType):
    def __init__(self, W, b, **kwargs):
        super().__init__(**kwargs)

        self.W = W
        self.b = b

        self.x = None
        self.original_x_shape = None
        # 가중치와 편향 매개변수의 미분
        self.dW = None
        self.db = None

    def forward(self, x):
        # 텐서 대응

        self.original_x_shape = x.shape

        x = x.reshape(x.shape[0], -1)

        self.x = x

        # print(self.x.shape, self.W.shape)

        out = np.dot(self.x, self.W) + self.b

        return out

    def backward(self, dout):
        dx = np.dot(dout, self.W.T)
        self.dW = np.dot(self.x.T, dout)
        self.db = np.sum(dout, axis=0)

        dx = dx.reshape(*self.original_x_shape)  # 입력 데
        return dx


class SigmoidWithLoss(LayerType):
    def __init__(self):

        super().__init__()

        self.y = None
        self.activation = True
        self.loss = None
        self.t = None

    def forward(self, x, t):
        self.t = t
        self.y = sigmoid(x)
        self.loss = cross_entropy_error()
        return self.loss

    def backward(self, dout):
        # dx = dout * (1.0 - self.out) * self.out

        # return dx
        batch_size = self.t.shape[0]
        if self.t.size == self.y.size:  # 정답 레이블이 원-핫 인코딩 형태일 때
            dx = (self.y - self.t) / batch_size
        else:
            dx = self.y.copy()
            dx[np.arange(batch_size), self.t] -= 1
            dx = dx / batch_size

        return dx



class SoftmaxWithLoss(LayerType):
    def __init__(self):

        super().__init__()


        self.name = "softmax"

        self.loss = None  # 손실함수
        self.y = None  # softmax의 출력
        self.t = None  # 정답 레이블(원-핫 인코딩 형태)

        self.activation = True

    def forward(self, x, t):
        self.t = t
        self.y = softmax(x)
        self.loss = cross_entropy_error(self.y, self.t)

        return self.loss

    def backward(self, dout=1):
        batch_size = self.t.shape[0]
        if self.t.size == self.y.size:  # 정답 레이블이 원-핫 인코딩 형태일 때
            dx = (self.y - self.t) / batch_size
        else:
            dx = self.y.copy()
            dx[np.arange(batch_size), self.t] -= 1
            dx = dx / batch_size

        return dx


class IdentityWithLoss(LayerType):
    def __init__(self):

        super().__init__()

        self.x = None
        self.t = None
        self.y = None
        self.loss = None
        self.activation = True

    def forward(self, x, t):
        self.x = x
        self.y = x
        self.t = t
        self.loss = np.square(x - self.t).mean()
        return self.loss

    def backward(self, dout=1):
        #         print(self.y, self.t)

        batch_size = self.t.shape[0]
        if self.t.size == self.y.size:  # 정답 레이블이 원-핫 인코딩 형태일 때
            #             dx = (self.y - self.t) / batch_size
            dx = (self.y - self.t)
        else:
            dx = self.y.copy()
            dx[np.arange(batch_size), self.t] -= 1
            dx = dx / batch_size

        return dx


class BatchNormalization(LayerType):
    """
    http://arxiv.org/abs/1502.03167
    """

    def __init__(self, gamma=None, beta=None, momentum=0.9, running_mean=None, running_var=None, **kwargs):

        super().__init__(**kwargs)

        self.gamma = gamma
        self.beta = beta
        self.momentum = momentum
        self.input_shape = None  # 합성곱 계층은 4차원, 완전연결 계층은 2차원

        # 시험할 때 사용할 평균과 분산
        self.running_mean = running_mean
        self.running_var = running_var

        # backward 시에 사용할 중간 데이터
        self.batch_size = None
        self.xc = None
        self.std = None
        self.dgamma = None
        self.dbeta = None

    def forward(self, x, train_flg=True):
        self.input_shape = x.shape
        if x.ndim != 2:
            N, C, H, W = x.shape
            x = x.reshape(N, -1)

        out = self.__forward(x, train_flg)

        return out.reshape(*self.input_shape)

    def __forward(self, x, train_flg):
        if self.running_mean is None:
            N, D = x.shape
            self.running_mean = np.zeros(D)
            self.running_var = np.zeros(D)

        if train_flg:
            mu = x.mean(axis=0)
            xc = x - mu
            var = np.mean(xc ** 2, axis=0)
            std = np.sqrt(var + 10e-7)
            xn = xc / std

            self.batch_size = x.shape[0]
            self.xc = xc
            self.xn = xn
            self.std = std

            if self.trainable is True:
                self.running_mean = self.momentum * self.running_mean + (1 - self.momentum) * mu
                self.running_var = self.momentum * self.running_var + (1 - self.momentum) * var
        else:
            xc = x - self.running_mean
            xn = xc / ((np.sqrt(self.running_var + 10e-7)))

        out = self.gamma * xn + self.beta
        return out

    def backward(self, dout):
        if dout.ndim != 2:
            N, C, H, W = dout.shape
            dout = dout.reshape(N, -1)

        dx = self.__backward(dout)

        dx = dx.reshape(*self.input_shape)
        return dx

    def __backward(self, dout):
        dbeta = dout.sum(axis=0)
        dgamma = np.sum(self.xn * dout, axis=0)
        dxn = self.gamma * dout
        dxc = dxn / self.std
        dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
        dvar = 0.5 * dstd / self.std
        dxc += (2.0 / self.batch_size) * self.xc * dvar
        dmu = np.sum(dxc, axis=0)
        dx = dxc - dmu / self.batch_size

        self.dgamma = dgamma
        self.dbeta = dbeta

        return dx


class Dropout(LayerType):
    """
    http://arxiv.org/abs/1207.0580
    """

    def __init__(self, dropout_ratio=0.5):

        super().__init__()

        self.dropout_ratio = dropout_ratio
        self.mask = None

    def forward(self, x, train_flg=True):
        if train_flg:
            self.mask = np.random.rand(*x.shape) > self.dropout_ratio
            return x * self.mask
        else:
            return x * (1.0 - self.dropout_ratio)

    def backward(self, dout):
        return dout * self.mask


class Conv2D(LayerType):

    def __init__(self, filter_num, filter_size, **kwargs):
        # print("ffff", kwargs)
        super().__init__(**kwargs)

        self.filter_num = filter_num
        self.filter_size = filter_size

        if "input_size" in kwargs:
            self.input_size = kwargs["input_size"]
        else:
            self.input_size = None
        if "activation" in kwargs:
            self.activation = kwargs["activation"]
        else:
            self.activation = None

        if "initializer" in kwargs:
            self.initializer = kwargs["initializer"]
        else:
            self.initializer = Initializer.Std(0.01)

        if "pad" in kwargs:
            self.pad = kwargs["pad"]
        else:
            self.pad = 0

        if "stride" in kwargs:
            self.stride = kwargs["stride"]
        else:
            self.stride = 1


class Convolution(LayerType):
    def __init__(self, W, b, stride=1, pad=0, **kwargs):
        # print(kwargs)
        super().__init__(**kwargs)

        self.W = W
        self.b = b
        self.stride = stride
        self.pad = pad

        # 중간 데이터（backward 시 사용）
        self.x = None
        self.col = None
        self.col_W = None

        # 가중치와 편향 매개변수의 기울기
        self.dW = None
        self.db = None

    def forward(self, x):

        # print(x.shape)

        FN, C, FH, FW = self.W.shape

        N, C, H, W = x.shape

        out_h = 1 + int((H + 2 * self.pad - FH) / self.stride)

        out_w = 1 + int((W + 2 * self.pad - FW) / self.stride)

        col = im2col(x, FH, FW, self.stride, self.pad)

        col_W = self.W.reshape(FN, -1).T

        out = np.dot(col, col_W) + self.b
        out = out.reshape(N, out_h, out_w, -1).transpose(0, 3, 1, 2)

        self.x = x
        self.col = col
        self.col_W = col_W

        return out

    def backward(self, dout):
        FN, C, FH, FW = self.W.shape
        dout = dout.transpose(0, 2, 3, 1).reshape(-1, FN)

        self.db = np.sum(dout, axis=0)
        self.dW = np.dot(self.col.T, dout)
        self.dW = self.dW.transpose(1, 0).reshape(FN, C, FH, FW)

        dcol = np.dot(dout, self.col_W.T)
        dx = col2im(dcol, self.x.shape, FH, FW, self.stride, self.pad)

        return dx


class Pooling(LayerType):
    def __init__(self, pool_h, pool_w, stride=1, pad=0):
        super().__init__()

        self.pool_h = pool_h
        self.pool_w = pool_w
        self.stride = stride
        self.pad = pad

        self.x = None
        self.arg_max = None

    def forward(self, x):
        N, C, H, W = x.shape
        out_h = int(1 + (H - self.pool_h) / self.stride)
        out_w = int(1 + (W - self.pool_w) / self.stride)

        col = im2col(x, self.pool_h, self.pool_w, self.stride, self.pad)
        col = col.reshape(-1, self.pool_h * self.pool_w)

        arg_max = np.argmax(col, axis=1)
        out = np.max(col, axis=1)
        out = out.reshape(N, out_h, out_w, C).transpose(0, 3, 1, 2)

        self.x = x
        self.arg_max = arg_max

        return out

    def backward(self, dout):
        dout = dout.transpose(0, 2, 3, 1)

        pool_size = self.pool_h * self.pool_w
        dmax = np.zeros((dout.size, pool_size))
        dmax[np.arange(self.arg_max.size), self.arg_max.flatten()] = dout.flatten()
        dmax = dmax.reshape(dout.shape + (pool_size,))

        dcol = dmax.reshape(dmax.shape[0] * dmax.shape[1] * dmax.shape[2], -1)
        dx = col2im(dcol, self.x.shape, self.pool_h, self.pool_w, self.stride, self.pad)

        return dx

# %%






import sys

import matplotlib.pyplot as plt
from jjy.framework.functions import *

sys.path.append("../../")

#
import cupy as np

import pickle
with open("train_data_2021-04-10 213537.pickle","rb") as fr:
    result = pickle.load(fr)


color_list = ["red", "blue", "green", "yellow", "purple"]

marker_list = ["o", "s", "^", "v","x" ]

plt.figure(figsize=(9, 5))


def show_loss():

    plt.xlabel("step")
    plt.ylabel("loss")

    x = np.arange(len(result["train_loss_list"]))
    plt.plot(x, result["train_loss_list"], label="loss",color = color_list[0],  marker=None)

    plt.legend(loc='upper right')

    plt.show()


def show_acc():
    plt.xlabel("epoch")
    plt.ylabel("acc")

    x = np.arange(len(result["train_acc_list"]))
    plt.plot(x, result["train_acc_list"], label="train_acc", color=color_list[0], marker=None)
    plt.plot(x, result["test_acc_list"], label="test_acc", color=color_list[1], marker="o")

    for i, v in enumerate(result["test_acc_list"]):
        plt.text(i, v + 0.0003, v,  # 좌표 (x축 = v, y축 = y[0]..y[1], 표시 = y[0]..y[1])
                 fontsize=9,
                 color='blue',
                 horizontalalignment='center',  # horizontalalignment (left, center, right)
                 verticalalignment='bottom')
    plt.legend(loc='lower right')

    plt.show()

#
show_loss()
show_acc()




import sys
import jjy.framework.initializer as Initializer
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
# from functions import *
from jjy.framework.network import MultiLayerNet

from jjy.dataset.mnist import load_mnist


def main():
    (x_train, t_train), (x_test, t_test) = load_mnist(flatten=False)
    print(x_train.shape, t_train.shape)
    print(t_train[0])

    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(128, (3, 3), pad=1, input_size=(1, 28, 28), trainable=True, ),
                  initializer=Initializer.He(),
                  activation=Layer.Relu())
    # net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Conv2D(32, (3, 3), pad=2, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Dense(50, initializer=Initializer.He(), activation=Layer.Relu()))
    # net.add_layer(Layer.Dropout(0.5))

    net.add_layer(Layer.FreezePrev())
    net.add_layer(Layer.Dense(10, trainable=True))

    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.SoftmaxWithLoss())

    for k, v in net.params.items():
        print(k, v.shape)

    result = net.train(
        x_train, t_train, x_test, t_test, batch_size=200, iters_num=3, print_epoch=1, evaluate_limit=100,
        is_use_progress_bar=True,
        optimizer=Optimizer.Adam(lr=0.001))

    import pickle
    import datetime
    # Save pickle

    # with open(f"train_data_{str(datetime.datetime.now())[:-7].replace(':', '')}.pickle", "wb") as fw:
        # pickle.dump(result, fw)
    # net.save_model()

    print("============================================")


main()

print("done!")




# %%

from jjy.framework.cuda import xp as np


def identity_function(x):
    return x


def step_function(x):
    return np.array(x > 0, dtype=np.int)


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def sigmoid_grad(x):
    return (1.0 - sigmoid(x)) * sigmoid(x)


def relu(x):
    return np.maximum(0, x)


def relu_grad(x):
    grad = np.zeros(x)
    grad[x >= 0] = 1
    return grad


def softmax(x):
    if x.ndim == 2:
        x = x.T
        x = x - np.max(x, axis=0)
        y = np.exp(x) / np.sum(np.exp(x), axis=0)
        return y.T

    x = x - np.max(x)  # 오버플로 대책
    return np.exp(x) / np.sum(np.exp(x))


def mean_squared_error(y, t):
    return 0.5 * np.sum((y - t) ** 2)


def cross_entropy_error(y, t):
    if y.ndim == 1:
        t = t.reshape(1, t.size)
        y = y.reshape(1, y.size)

    # 훈련 데이터가 원-핫 벡터라면 정답 레이블의 인덱스로 반환
    if t.size == y.size:
        t = t.argmax(axis=1)

    batch_size = y.shape[0]


    return -np.sum(np.log(y[np.arange(batch_size), t] + 1e-7)) / batch_size


def softmax_loss(X, t):
    y = softmax(X)
    return cross_entropy_error(y, t)


def shuffle_split_data(X, y, ratio=0.2):
    arr_rand = np.random.rand(X.shape[0])
    split = arr_rand < np.percentile(arr_rand, (1 - ratio) * 100)

    X_train = X[split]
    y_train = y[split]
    X_test = X[~split]
    y_test = y[~split]

    return X_train, y_train, X_test, y_test


def conv_output_size(input_size, filter_size, stride=1, pad=0):
    return (input_size + 2 * pad - filter_size) / stride + 1


def im2col(input_data, filter_h, filter_w, stride=1, pad=0):
    """다수의 이미지를 입력받아 2차원 배열로 변환한다(평탄화).

    Parameters
    ----------
    input_data : 4차원 배열 형태의 입력 데이터(이미지 수, 채널 수, 높이, 너비)
    filter_h : 필터의 높이
    filter_w : 필터의 너비
    stride : 스트라이드
    pad : 패딩

    Returns
    -------
    col : 2차원 배열
    """
    N, C, H, W = input_data.shape
    out_h = (H + 2 * pad - filter_h) // stride + 1
    out_w = (W + 2 * pad - filter_w) // stride + 1

    img = np.pad(input_data, [(0, 0), (0, 0), (pad, pad), (pad, pad)], 'constant')
    col = np.zeros((N, C, filter_h, filter_w, out_h, out_w))

    for y in range(filter_h):
        y_max = y + stride * out_h
        for x in range(filter_w):
            x_max = x + stride * out_w
            col[:, :, y, x, :, :] = img[:, :, y:y_max:stride, x:x_max:stride]

    col = col.transpose(0, 4, 5, 1, 2, 3)
    # print(N * out_h * out_w)
    col = col.reshape(N * out_h * out_w, -1)
    return col


def col2im(col, input_shape, filter_h, filter_w, stride=1, pad=0):
    """(im2col과 반대) 2차원 배열을 입력받아 다수의 이미지 묶음으로 변환한다.

    Parameters
    ----------
    col : 2차원 배열(입력 데이터)
    input_shape : 원래 이미지 데이터의 형상（예：(10, 1, 28, 28)）
    filter_h : 필터의 높이
    filter_w : 필터의 너비
    stride : 스트라이드
    pad : 패딩

    Returns
    -------
    img : 변환된 이미지들
    """
    N, C, H, W = input_shape
    out_h = (H + 2 * pad - filter_h) // stride + 1
    out_w = (W + 2 * pad - filter_w) // stride + 1
    col = col.reshape(N, out_h, out_w, C, filter_h, filter_w).transpose(0, 3, 4, 5, 1, 2)

    img = np.zeros((N, C, H + 2 * pad + stride - 1, W + 2 * pad + stride - 1))
    for y in range(filter_h):
        y_max = y + stride * out_h
        for x in range(filter_w):
            x_max = x + stride * out_w
            img[:, :, y:y_max:stride, x:x_max:stride] += col[:, :, y, x, :, :]

    return img[:, :, pad:H + pad, pad:W + pad]

# %%



def make_sample_data_set():
    """
        x는 0~999999 정수이다.
        정답은 x를 100000으로 나눴을 때의 몫이다.
    """

    x = np.random.randint(999999, size=(10000, 1))

    t_data = x.flatten() // 100000

    # t_data 원핫 인코딩 코드
    num = np.unique(t_data, axis=0)
    num = num.shape[0]
    t = np.eye(num)[t_data]

    return x, t


def make_sample_data_set_regression():
    x = np.random.randint(9, size=(300, 1))
    y = np.dot(x, np.array([3]))
    t = np.reshape(y, (y.shape[0], 1))

    return x, t


def make_sample_data_set_regression2():
    x = np.random.randint(999999, size=(300, 2))
    y = np.dot(x, np.array([3, -2])) + 7
    #     y =  y + (4 * np.random.random_sample((1,y.shape[0])) - 2).flatten()
    t = np.reshape(y, (y.shape[0], 1))

    return x, t

def foo():
    print("hello")
    return


# %%

# coding: utf-8
try:
    import urllib.request
except ImportError:
    raise ImportError('You should use Python 3.x')
import os.path
import gzip
import pickle
import os
from jjy.framework.cuda import xp as np

url_base = 'http://yann.lecun.com/exdb/mnist/'
key_file = {
    'train_img': 'train-images-idx3-ubyte.gz',
    'train_label': 'train-labels-idx1-ubyte.gz',
    'test_img': 't10k-images-idx3-ubyte.gz',
    'test_label': 't10k-labels-idx1-ubyte.gz'
}

# dataset_dir = os.path.dirname(os.path.abspath("__file__"))
dataset_dir = os.path.dirname(os.path.realpath(__file__))

save_file = dataset_dir + "/mnist.pkl"

train_num = 60000
test_num = 10000
img_dim = (1, 28, 28)
img_size = 784


def _download(file_name):
    file_path = dataset_dir + "/" + file_name
    print(file_path)

    if os.path.exists(file_path):
        return

    print("Downloading " + file_name + " ... ")
    urllib.request.urlretrieve(url_base + file_name, file_path)
    print("Done")


def download_mnist():
    print(dataset_dir)
    for v in key_file.values():
        _download(v)


def _load_label(file_name):
    file_path = dataset_dir + "/" + file_name

    print("Converting " + file_name + " to NumPy Array ...")
    with gzip.open(file_path, 'rb') as f:
        labels = np.frombuffer(f.read(), np.uint8, offset=8)
    print("Done")

    return labels


def _load_img(file_name):
    file_path = dataset_dir + "/" + file_name

    print("Converting " + file_name + " to NumPy Array ...")
    with gzip.open(file_path, 'rb') as f:
        data = np.frombuffer(f.read(), np.uint8, offset=16)
    data = data.reshape(-1, img_size)
    print("Done")

    return data


def _convert_numpy():
    dataset = {}
    dataset['train_img'] = _load_img(key_file['train_img'])
    dataset['train_label'] = _load_label(key_file['train_label'])
    dataset['test_img'] = _load_img(key_file['test_img'])
    dataset['test_label'] = _load_label(key_file['test_label'])

    return dataset


def init_mnist():
    download_mnist()
    dataset = _convert_numpy()
    print("Creating pickle file ...")
    with open(save_file, 'wb') as f:
        pickle.dump(dataset, f, -1)
    print("Done!")


def _change_ont_hot_label(X):
    T = np.zeros((X.size, 10))
    for idx, row in enumerate(T):
        row[X[idx]] = 1

    return T


def load_mnist(normalize=True, flatten=True, one_hot_label=False):
    """MNIST 데이터셋 읽기

    Parameters
    ----------
    normalize : 이미지의 픽셀 값을 0.0~1.0 사이의 값으로 정규화할지 정한다.
    one_hot_label :
        one_hot_label이 True면、레이블을 원-핫(one-hot) 배열로 돌려준다.
        one-hot 배열은 예를 들어 [0,0,1,0,0,0,0,0,0,0]처럼 한 원소만 1인 배열이다.
    flatten : 입력 이미지를 1차원 배열로 만들지를 정한다.

    Returns
    -------
    (훈련 이미지, 훈련 레이블), (시험 이미지, 시험 레이블)
    """
    #     print("hello")

    if not os.path.exists(save_file):
        print(save_file)
        init_mnist()


    with open(save_file, 'rb') as f:
        dataset = pickle.load(f)

    if normalize:
        for key in ('train_img', 'test_img'):
            dataset[key] = dataset[key].astype(np.float32)
            dataset[key] /= 255.0

    if one_hot_label:
        dataset['train_label'] = _change_ont_hot_label(dataset['train_label'])
        dataset['test_label'] = _change_ont_hot_label(dataset['test_label'])

    if not flatten:
        for key in ('train_img', 'test_img'):
            dataset[key] = dataset[key].reshape(-1, 1, 28, 28)

    return (dataset['train_img'], dataset['train_label']), (dataset['test_img'], dataset['test_label'])


if __name__ == '__main__':
    init_mnist()

# %%

os.path.dirname(os.path.abspath("__file__"))

# %%





import sys

import matplotlib.pyplot as plt
from jjy.framework.functions import *
import matplotlib.pyplot as plt
from jjy.framework.functions import *
from PIL import Image
import os
#
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
import jjy.framework.initializer as Initializer
from jjy.framework.functions import *
from jjy.framework.network import MultiLayerNet
import random
import cupy as cp
import numpy as np2


fpath = "./idol/idol_result/idol_train_weight_2021-04-20 054113_8381.npz"
output_path = "./idol/idol_result/idol_train_weight_2021-04-20 054113_8381_np.npz"

load_data = dict(np2.load(fpath, allow_pickle=True))

print(type(load_data["Params"]))

print(load_data["Params"].item()["b18"])

for k, v in load_data.items():
    if k == "OtherParams" or k == "Params":
        for param_name in load_data.get(k).item():
            load_data[k].item()[param_name] = cp.asnumpy(load_data[k].item()[param_name])

np.savez_compressed(output_path, **load_data)

# print(load_data)


# %%

import sys

import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
import jjy.framework.initializer as Initializer
from jjy.framework.functions import *
from jjy.framework.network import MultiLayerNet
import jjy.framework.scaler as Scaler
from jjy.dataset.mnist import load_mnist

from jjy.framework.cuda import xp as np


def make_sample_data_set():
    """
        x는 0~999999 정수이다.
        정답은 x를 100000으로 나눴을 때의 몫이다.
    """

    x = np.random.randint(999999, size=(10000, 1))

    t_data = x.flatten() // 100000

    # t_data 원핫 인코딩 코드
    num = np.unique(t_data, axis=0)
    num = num.shape[0]
    t = np.eye(num)[t_data]

    return x, t


def make_sample_data_set_regression():
    x = np.random.randint(3 , size=(300, 1))
    y = np.dot(x, np.array([2]))
    t = np.reshape(y, (y.shape[0], 1))

    return x, t


def make_sample_data_set_regression2():
    x = np.random.randint(999999, size=(300, 2))
    y = np.dot(x, np.array([3, -2])) + 7
    #     y =  y + (4 * np.random.random_sample((1,y.shape[0])) - 2).flatten()
    t = np.reshape(y, (y.shape[0], 1))

    return x, t

def make_sample_data_set_regression3():
    x = np.random.rand(300, 2) * 3
    y = np.dot(x, np.array([2, 1])) + 3
    #     y =  y + (4 * np.random.random_sample((1,y.shape[0])) - 2).flatten()
    t = np.reshape(y, (y.shape[0], 1))

    # print(x)

    return x, t



x_data, t_data = make_sample_data_set_regression3()

print(x_data[:3])
print(t_data[:3])

net = MultiLayerNet()
net.add_layer(Layer.Dense(1, input_size = 2, activation=Layer.IdentityWithLoss() ))
# net.add_layer(Layer.Dense(5, input_size = 2, activation=Layer.Relu() ))
# net.add_layer(Layer.Dense(1))

x_train, t_train, x_test, t_test = shuffle_split_data(x_data, t_data, 0.2)

print(net.params)

# scaler = Scaler.StandardScaler()
# x_train = scaler.fit_transform(x_train)
# x_test = scaler.transform(x_test)

result = net.train(
        x_train, t_train, x_test, t_test, batch_size = 100, iters_num = 1000, print_epoch = 30,
        optimizer = Optimizer.SGD(lr=0.001)
)


print("done!")

print(net.params)
print(net.layers["Affine0"].__dict__)


net.save_model()

# net.load_model("weight.npz")


from jjy.framework.cuda import xp as np
import jjy.framework.cuda


import sys
import jjy.framework.initializer as Initializer
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
# from functions import *
from jjy.framework.network import MultiLayerNet

from jjy.dataset.mnist import load_mnist

from jjy.framework.cuda import GPU_ENABLE


def main():
    (x_train, t_train), (x_test, t_test) = load_mnist(flatten=False)
    print(x_train.shape, t_train.shape)
    # print(t_train[0])

    # jjy.framework.cuda.disable_gpu()
    # print(np, type(np))

    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, input_size=(1, 28, 28)), initializer=Initializer.He(),
                  activation=Layer.Relu())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dense(20, initializer=Initializer.He(), activation=Layer.Relu()))
    net.add_layer(Layer.Dense(10))
    net.add_layer(Layer.BatchNormalization())

    net.add_layer(Layer.SoftmaxWithLoss())




    for k, v in net.params.items():
        print(k, v.shape)

    result = net.train(
        x_train, t_train, x_test, t_test, batch_size=100, iters_num=5000, print_epoch=1, evaluate_limit=1000,
        is_use_progress_bar=True,
        optimizer=Optimizer.Adam(lr=0.001))

    import pickle
    import datetime
    ## Save pickle
    with open(f"train_data_{str(datetime.datetime.now())[:-7].replace(':', '')}.pickle", "wb") as fw:
        pickle.dump(result, fw)
    net.save_model()

    print("============================================")


main()

print("done!")



# %%

from jjy.framework.cuda import xp as np


class OptimizerType:

    def __init__(self, **kwargs):
        if "trainable_param_keys" in kwargs:
            self.trainable_param_keys = kwargs["trainable_param_keys"]
        else:
            self.trainable_param_keys = None


    def get_trainable_param_keys(self, param_keys):
        if self.trainable_param_keys is None:
            return param_keys
        else:
            return self.trainable_param_keys


class SGD(OptimizerType):
    """확률적 경사 하강법（Stochastic Gradient Descent）"""

    def __init__(self, lr=0.01, **kwargs):
        super().__init__(**kwargs)

        self.lr = lr

    def update(self, params, grads):
        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            params[key] -= self.lr * grads[key]


class Momentum(OptimizerType):
    """모멘텀 SGD"""

    def __init__(self, lr=0.01, momentum=0.9, **kwargs):
        super().__init__(**kwargs)
        self.lr = lr
        self.momentum = momentum
        self.v = None

    def update(self, params, grads):
        if self.v is None:
            self.v = {}
            for key, val in params.items():
                self.v[key] = np.zeros_like(val)

        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            self.v[key] = self.momentum * self.v[key] - self.lr * grads[key]
            params[key] += self.v[key]


class Nesterov(OptimizerType):
    """Nesterov's Accelerated Gradient (http://arxiv.org/abs/1212.0901)"""

    # NAG는 모멘텀에서 한 단계 발전한 방법이다. (http://newsight.tistory.com/224)

    def __init__(self, lr=0.01, momentum=0.9, **kwargs):
        super().__init__(**kwargs)
        self.lr = lr
        self.momentum = momentum
        self.v = None

    def update(self, params, grads):
        if self.v is None:
            self.v = {}
            for key, val in params.items():
                self.v[key] = np.zeros_like(val)

        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            self.v[key] *= self.momentum
            self.v[key] -= self.lr * grads[key]
            params[key] += self.momentum * self.momentum * self.v[key]
            params[key] -= (1 + self.momentum) * self.lr * grads[key]


class AdaGrad(OptimizerType):
    """AdaGrad"""

    def __init__(self, lr=0.01, **kwargs):
        super().__init__(**kwargs)
        self.lr = lr
        self.h = None

    def update(self, params, grads):
        if self.h is None:
            self.h = {}
            for key, val in params.items():
                self.h[key] = np.zeros_like(val)

        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            self.h[key] += grads[key] * grads[key]
            params[key] -= self.lr * grads[key] / (np.sqrt(self.h[key]) + 1e-7)


class RMSprop(OptimizerType):
    """RMSprop"""

    def __init__(self, lr=0.01, decay_rate=0.99, **kwargs):
        super().__init__(**kwargs)
        self.lr = lr
        self.decay_rate = decay_rate
        self.h = None

    def update(self, params, grads):
        if self.h is None:
            self.h = {}
            for key, val in params.items():
                self.h[key] = np.zeros_like(val)

        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            self.h[key] *= self.decay_rate
            self.h[key] += (1 - self.decay_rate) * grads[key] * grads[key]
            params[key] -= self.lr * grads[key] / (np.sqrt(self.h[key]) + 1e-7)


class Adam(OptimizerType):
    """Adam (http://arxiv.org/abs/1412.6980v8)"""

    def __init__(self, lr=0.001, beta1=0.9, beta2=0.999, **kwargs):
        super().__init__(**kwargs)
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.iter = 0
        self.m = None
        self.v = None

    def update(self, params, grads):
        if self.m is None:
            self.m, self.v = {}, {}
            for key, val in params.items():
                self.m[key] = np.zeros_like(val)
                self.v[key] = np.zeros_like(val)

        self.iter += 1
        lr_t = self.lr * np.sqrt(1.0 - self.beta2 ** self.iter) / (1.0 - self.beta1 ** self.iter)

        for key in self.get_trainable_param_keys(param_keys=params.keys()):
            # self.m[key] = self.beta1*self.m[key] + (1-self.beta1)*grads[key]
            # self.v[key] = self.beta2*self.v[key] + (1-self.beta2)*(grads[key]**2)
            self.m[key] += (1 - self.beta1) * (grads[key] - self.m[key])
            self.v[key] += (1 - self.beta2) * (grads[key] ** 2 - self.v[key])

            params[key] -= lr_t * self.m[key] / (np.sqrt(self.v[key]) + 1e-7)

            # unbias_m += (1 - self.beta1) * (grads[key] - self.m[key]) # correct bias
            # unbisa_b += (1 - self.beta2) * (grads[key]*grads[key] - self.v[key]) # correct bias
            # params[key] += self.lr * unbias_m / (np.sqrt(unbisa_b) + 1e-7)



import os

GPU_ENABLE = os.environ.get("JJY_GPU_ENABLE", "True")
# GPU_ENABLE = "False"

GPU_MODE = True
if GPU_ENABLE.lower() == "false":
    GPU_ENABLE = False
else:
    GPU_ENABLE = True

import numpy as np

xp = np


if GPU_ENABLE is True:

    try:
        import cupy as cp

        xp = cp
        xp.cuda.Device(0).use()
        GPU_MODE = True
    except ImportError:
        GPU_ENABLE = False
        GPU_MODE = False

# a = [1,2,3]
# # # a = list(a)
# b = xp.prod(np.asarray(a))
# print(b)
# c = xp.ones(int(b))
# print(c)

# def enable_gpu():
#     global xp
#     try:
#         import cupy as cp
#
#         xp = cp
#     except ImportError:
#
#         gpu_enable = False
#         raise ImportError("Cupy 설치해주세요~~")
#
#
# def disable_gpu():
#     global xp
#     try:
#         import numpy as np
#         xp = np
#     except ImportError:
#
#         gpu_enable = False
#         raise ImportError("Numpy 설치해주세요~~")



def get_array_module(x):
    """Returns the array module for `x`.
    Args:
        x (dezero.Variable or numpy.ndarray or cupy.ndarray): Values to
            determine whether NumPy or CuPy should be used.
    Returns:
        module: `cupy` or `numpy` is returned based on the argument.
    """

    if not GPU_ENABLE:
        return np
    xp = cp.get_array_module(x)
    return xp


def as_numpy(x):
    """Convert to `numpy.ndarray`.
    Args:
        x (`numpy.ndarray` or `cupy.ndarray`): Arbitrary object that can be
            converted to `numpy.ndarray`.
    Returns:
        `numpy.ndarray`: Converted array.
    """

    if np.isscalar(x):
        return np.array(x)
    elif isinstance(x, np.ndarray):
        return x
    return cp.asnumpy(x)


def as_cupy(x):
    """Convert to `cupy.ndarray`.
    Args:
        x (`numpy.ndarray` or `cupy.ndarray`): Arbitrary object that can be
            converted to `cupy.ndarray`.
    Returns:
        `cupy.ndarray`: Converted array.
    """

    if not GPU_ENABLE:
        raise Exception('CuPy cannot be loaded. Install CuPy!')
    return cp.asarray(x)



# %%

from jjy.framework.cuda import xp as np


# %%

class StandardScaler:
    def __init__(self):
        self.std_array = None
        self.mean_array = None

    def fit(self, data):
        self.data = data

        self.std_array = np.std(data, axis=0)
        self.mean_array = np.mean(data, axis=0)

    def transform(self, data):
        return (data - self.mean_array) / self.std_array

    def fit_transform(self, data):
        self.fit(data)
        return self.transform(data)

# %%





import os
from PIL import Image

import matplotlib.pyplot as plt
import numpy as np


from tqdm import tqdm
import random





base_dir = 'D:\\google-image-crawler\\dataset'

train_dir = os.path.join(base_dir, 'train_set')
test_dir = os.path.join(base_dir, 'test_set')

idol_list = ["iu", "irene", "arin"]

x_train = []
t_train = []

x_test = []
t_test = []


def img_to_array(fname):
    image = Image.open(fname)

    image = image.resize((128, 128))
    # image = np.reshape(image, (, 128, 128))
    # image = np.reshape(image.convert("L"), (1, 128, 128))
    # show_img_by_array(np.asarray(image))
    # return np.asarray(image)
    return np.transpose(np.asarray(image), (2, 1, 0))


def show_img_by_array(img_array):
    transposed_array = np.transpose(img_array, (1, 2, 0))
    plt.imshow(transposed_array, cmap='gray')
    plt.show()


for idx, idol in enumerate(idol_list):


    train_idol_dir = os.path.join(train_dir, f'{idol}\\output')
    test_idol_dir = os.path.join(test_dir, f'{idol}')

    train_idol_fname_list = os.listdir(train_idol_dir)
    test_idol_fname_list = os.listdir(test_idol_dir)

    random.shuffle(train_idol_fname_list)

    for fname in tqdm(train_idol_fname_list[:5000]):
        x_train.append(img_to_array(os.path.join(train_idol_dir, fname)))
        t_train.append(idx)


    for fname in tqdm(test_idol_fname_list):
        x_test.append(img_to_array(os.path.join(test_idol_dir, fname)))
        t_test.append(idx)

x_train = np.array(x_train)
x_test = np.array(x_test)
t_train = np.array(t_train)
t_test = np.array(t_test)

print(x_train.shape)
temp_data = {"x_train" : x_train, "x_test" : x_test, "t_train" : t_train, "t_test" : t_test}
np.savez_compressed("./idol_images_rgb_128_v2.npz", **temp_data)


import os
import shutil
import random

base_dir = 'D:\\google-image-crawler'

crop_dir = os.path.join(base_dir, 'crop')

idol_list = ["iu", "irene", "arin"]
for idol in idol_list:

    idol_dir = os.path.join(crop_dir, idol)
    fname_list = os.listdir(idol_dir)

    random.shuffle(fname_list)

    for fname in fname_list[:1000]:
        shutil.copy(os.path.join(idol_dir, fname), f"D:\\google-image-crawler\\dataset\\train_set\\{idol}")
    for fname in fname_list[1000:]:
        shutil.copy(os.path.join(idol_dir, fname), f"D:\\google-image-crawler\\dataset\\test_set\\{idol}")






import sys

import matplotlib.pyplot as plt
from jjy.framework.functions import *
from PIL import Image
import os
#
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
import jjy.framework.initializer as Initializer
from jjy.framework.functions import *
from jjy.framework.network import MultiLayerNet
import random

import cupy as np

import pickle
with open("train_data_2021-04-17 061246.pickle","rb") as fr:
    result = pickle.load(fr)


color_list = ["red", "blue", "green", "yellow", "purple"]

marker_list = ["o", "s", "^", "v","x" ]

plt.figure(figsize=(9, 5))


def show_loss():

    plt.xlabel("step")
    plt.ylabel("loss")

    x = np.arange(len(result["train_loss_list"]))
    plt.plot(x, result["train_loss_list"], label="loss",color = color_list[0],  marker=None)

    plt.legend(loc='upper right')

    plt.show()


def show_acc():
    plt.xlabel("epoch")
    plt.ylabel("acc")

    x = np.arange(len(result["train_acc_list"]))
    plt.plot(x, result["train_acc_list"], label="train_acc", color=color_list[0], marker=None)
    plt.plot(x, result["test_acc_list"], label="test_acc", color=color_list[1], marker="o")

    for i, v in enumerate(result["test_acc_list"]):
        plt.text(i, v + 0.0003, v,  # 좌표 (x축 = v, y축 = y[0]..y[1], 표시 = y[0]..y[1])
                 fontsize=9,
                 color='blue',
                 horizontalalignment='center',  # horizontalalignment (left, center, right)
                 verticalalignment='bottom')
    plt.legend(loc='lower right')

    plt.show()

def show_img(img,t, ax):
    ax.grid(False)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.imshow(img)
    ax.set_title(t, fontsize=20)


def plot_grid(imgs, title_list, nrows, ncols, figsize=(10, 10)):



    assert len(imgs) == nrows * ncols, f"Number of images should be {nrows}x{ncols}"
    _, axs = plt.subplots(nrows, ncols, figsize=figsize)
    axs = axs.flatten()
    for i, (img, ax) in enumerate(zip(imgs, axs)):
        show_img(img,title_list[i], ax)


def show_img_predict():
    net = MultiLayerNet()
    net.load_model("dog_cat_f.npz")

    print(net.layers)
    print(net.params)


    base_dir = 'C:\\Users\\User\\deep-learning-without-tensorflow\\src\\jjy\\jjy\\dataset\\dog_cat'

    test_dir = os.path.join(base_dir, 'test_set\\test_set')

    # 테스트에 사용되는 고양이/개 이미지 경로
    test_cats_dir = os.path.join(test_dir, 'cats')
    test_dogs_dir = os.path.join(test_dir, 'dogs')
    print(test_cats_dir)
    print(test_dogs_dir)

    print('Total validation cat images :', len(os.listdir(test_cats_dir)))
    print('Total validation dog images :', len(os.listdir(test_dogs_dir)))

    test_cat_fnames = os.listdir(test_cats_dir)
    test_dog_fnames = os.listdir(test_dogs_dir)

    def img_to_array(fname, original=False):

        image = Image.open(fname)

        if original is False:
            image = image.resize((64, 64))
            image = np.reshape(image.convert("L"), (1, 64, 64))
        # show_img_by_array(np.asarray(image))
        return np.asarray(image)

    random.shuffle(test_cat_fnames)
    random.shuffle(test_dog_fnames)

    img_list = []
    predict_list = []
    for fname in test_dog_fnames[:15]:

        fname = os.path.join(test_dogs_dir, fname)

        img_original_array = img_to_array(fname, original=True)
        img_array = img_to_array(fname)
        img_array = img_array / 255.0

        img_list.append(img_original_array)
        predict_num = np.argmax(net.predict(np.array([img_array]), train_flg=False), axis=1)[0]
        print(net.predict(np.array([img_array]), train_flg=False))
        # print(predict_num)

        predict_list.append("CAT!" if predict_num == 0 else "DOG!")

    plot_grid(img_list, predict_list, 3, 5)
    plt.show()




show_img_predict()

# show_loss()
# show_acc()




# %%

from jjy.framework.functions import *
from jjy.framework.gradient import *


# %%

class InitializerType:
    pass


class Std(InitializerType):

    def __init__(self, std):
        self.std = std


class He(InitializerType):
    pass


class Xavier(InitializerType):
    pass







import os
from PIL import Image
import jsonpickle as jsp
import matplotlib.pyplot as plt
import cupy as np
import time
import random
from tqdm import tqdm
import jjy.framework.initializer as Initializer
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
# from functions import *
from jjy.framework.network import MultiLayerNet

from jjy.dataset.mnist import load_mnist
import datetime


class Timer(object):

    def __init__(self, name=""):
        self.name = name

    def __enter__(self):
        self.t = datetime.datetime.now()
        return self

    def __exit__(self, type, value, traceback):
        self.t = datetime.datetime.now() - self.t
        print(f"{self.name} Time : ", self.t)


# Open the image form working directory

def img_to_array(fname):
    image = Image.open(fname)
    image = image.resize((64, 64))
    image = np.reshape(image.convert("L"), (1, 64, 64))
    # show_img_by_array(np.asarray(image))
    return np.asarray(image)
    # return np.transpose(np.asarray(image), (2, 1, 0))


def show_img_by_array(img_array):
    transposed_array = np.transpose(img_array, (1, 2, 0))
    plt.imshow(transposed_array, cmap='gray')
    plt.show()


def get_images_from_dataset():
    base_dir = 'C:\\Users\\User\\deep-learning-without-tensorflow\\src\\jjy\\jjy\\dataset\\dog_cat'

    train_dir = os.path.join(base_dir, 'training_set\\training_set')
    test_dir = os.path.join(base_dir, 'test_set\\test_set')

    # 훈련에 사용되는 고양이/개 이미지 경로
    train_cats_dir = os.path.join(train_dir, 'cats\\output')
    train_dogs_dir = os.path.join(train_dir, 'dogs\\output')
    print(train_cats_dir)
    print(train_dogs_dir)

    # 테스트에 사용되는 고양이/개 이미지 경로
    test_cats_dir = os.path.join(test_dir, 'cats')
    test_dogs_dir = os.path.join(test_dir, 'dogs')
    print(test_cats_dir)
    print(test_dogs_dir)

    print('Total training cat images :', len(os.listdir(train_cats_dir)))
    print('Total training dog images :', len(os.listdir(train_dogs_dir)))

    print('Total validation cat images :', len(os.listdir(test_cats_dir)))
    print('Total validation dog images :', len(os.listdir(test_dogs_dir)))

    train_cat_fnames = os.listdir(train_cats_dir)
    train_dog_fnames = os.listdir(train_dogs_dir)

    test_cat_fnames = os.listdir(test_cats_dir)
    test_dog_fnames = os.listdir(test_dogs_dir)

    x_train = []

    for fname in tqdm(train_cat_fnames):
        x_train.append(img_to_array(os.path.join(train_cats_dir, fname)))

    for fname in tqdm(train_dog_fnames):
        x_train.append(img_to_array(os.path.join(train_dogs_dir, fname)))
    x_train = np.array(x_train)
    t_train = np.concatenate((np.zeros(len(train_cat_fnames)), np.ones(len(train_dog_fnames))), axis=0).astype(np.int64)

    x_test = []
    for fname in tqdm(test_cat_fnames):
        x_test.append(img_to_array(os.path.join(test_cats_dir, fname)))
    for fname in tqdm(test_dog_fnames):
        x_test.append(img_to_array(os.path.join(test_dogs_dir, fname)))
    x_test = np.array(x_test)
    t_test = np.concatenate((np.zeros(len(test_cat_fnames)), np.ones(len(test_dog_fnames))), axis=0).astype(np.int64)

    return x_train, t_train, x_test, t_test


# x_train, t_train, x_test, t_test = get_images_from_dataset()
# temp_data = {"x_train" : x_train, "x_test" : x_test, "t_train" : t_train, "t_test" : t_test}
# np.savez_compressed("./dog_cat_images_64.npz", **temp_data)
# exit(1)

def make_net1():
    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, input_size=(1, 64, 64)), initializer=Initializer.He())
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(128, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dense(50, initializer=Initializer.He(), activation=Layer.Relu()))
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Dense(2))
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.SoftmaxWithLoss())
    return net


def make_net2():
    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, input_size=(1, 64, 64)), initializer=Initializer.He(),
                  activation=Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dense(128, initializer=Initializer.He(), activation=Layer.Relu()))
    net.add_layer(Layer.Dense(2, initializer=Initializer.He()))
    net.add_layer(Layer.SoftmaxWithLoss())

    return net


# def make_net2():
#     net = MultiLayerNet(is_use_dropout=False)
#     net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, input_size=(1, 64, 64)), initializer=Initializer.He(),
#                   activation=Layer.Relu())
#     net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
#     net.add_layer(Layer.Dense(128, initializer=Initializer.He(), activation=Layer.Relu()))
#     net.add_layer(Layer.Dense(2, initializer=Initializer.He()))
#     net.add_layer(Layer.SoftmaxWithLoss())
#
#     return net


def shuffle_dataset(x, t):
    from sklearn.utils import shuffle
    x, t = shuffle(x, t)
    return x, t


load_data = np.load("./dog_cat_images_64.npz", allow_pickle=True)
x_train = load_data["x_train"]

t_train = load_data["t_train"].astype(np.int64)

x_test = load_data["x_test"]
t_test = load_data["t_test"].astype(np.int64)

x_train, t_train = shuffle_dataset(x_train, t_train)
x_test, t_test = shuffle_dataset(x_test, t_test)

x_train = x_train / 255
x_test = x_test / 255

# exit(1)

#
#
def train_model():
    net = make_net1()

    result = net.train(
        x_train, t_train, x_test, t_test, batch_size=64, iters_num=2000, print_epoch=1, evaluate_limit=100,
        is_use_progress_bar=True, save_model_each_epoch=1, save_model_path="./dog_cat_result",
        optimizer=Optimizer.Adam(lr=0.0005))

    # print(layer_info)
    #
    # print(layer_info[-2])
    # print(layer_info)
    # exit(1)

    import pickle
    import datetime
    #
    ## Save pickle
    with open(f"train_data_{str(datetime.datetime.now())[:-7].replace(':', '')}.pickle", "wb") as fw:
        pickle.dump(result, fw)
    net.save_model("dog_cat_f2.npz")

    print("============================================")


train_model()

def load_trained_model():
    net = MultiLayerNet()

    net.load_model("dog_cat_f.npz")

    print(net.params.keys())
    return

    for k, v in net.params.items():
        print(k, v.shape)

    for i in range(5):
        img_idx = random.randrange(0, 1000)
        transposed_array = np.transpose(x_test[img_idx], (1, 2, 0))
        plt.imshow(transposed_array, cmap='gray')
        plt.show()
        predict_num = np.argmax(net.predict(np.array([x_test[img_idx]]), train_flg=False), axis=1)[0]
        print(net.predict(np.array([x_test[img_idx]])))
        correct_num = t_test[img_idx]
        print(f"Predict : {predict_num}, Correct : {correct_num}")

# load_trained_model()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "matplotlib==3.4.1",
    'numpy==1.20.2',
    'tqdm==4.60.0',
    'jsonpickle==2.0.0'
]

setuptools.setup(
    name="jjy", # Replace with your own username
    version="0.0.3.1",
    author="jjy37777",
    author_email="jjy37777@naver.com",
    description="jjy ai framework",
    long_description=long_description,
    install_requires=install_requires,
    long_description_content_type="text/markdown",
    url="https://github.com/woduq1414/deep-learning-without-framework",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

# python setup.py sdist bdist_wheel


#python -m twine upload dist/*




from jjy.framework.network import MultiLayerNet
import random
from jjy.dataset.mnist import load_mnist
from jjy.framework.cuda import xp as np
import matplotlib.pyplot as plt

(x_train, t_train), (x_test, t_test) = load_mnist(flatten=False)

net = MultiLayerNet()
net.load_model("train_weight_2021-04-18 150634.npz")



for i in range(5):
    img_idx = random.randrange(0, 10000)

    plt.imshow(x_test[img_idx].reshape(28, 28), cmap='gray')
    plt.show()
    predict_num = np.argmax(net.predict(np.array([x_test[img_idx]]), train_flg=False), axis=1)[0]
    correct_num = t_test[img_idx]
    print(f"Predict : {predict_num}, Correct : {correct_num}")




import os
from PIL import Image
import jsonpickle as jsp
import matplotlib.pyplot as plt
from jjy.framework.cuda import xp as np
import numpy as np2

import time
import random
from tqdm import tqdm
import jjy.framework.initializer as Initializer
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
# from functions import *
from jjy.framework.network import MultiLayerNet

from jjy.dataset.mnist import load_mnist
import datetime


class Timer(object):

    def __init__(self, name=""):
        self.name = name

    def __enter__(self):
        self.t = datetime.datetime.now()
        return self

    def __exit__(self, type, value, traceback):
        self.t = datetime.datetime.now() - self.t
        print(f"{self.name} Time : ", self.t)


# Open the image form working directory

def img_to_array(fname):
    image = Image.open(fname)
    image = image.resize((64, 64))
    image = np.reshape(image.convert("L"), (1, 64, 64))
    # show_img_by_array(np.asarray(image))
    return np.asarray(image)
    # return np.transpose(np.asarray(image), (2, 1, 0))


def show_img_by_array(img_array):
    transposed_array = np.transpose(img_array, (1, 2, 0))
    plt.imshow(transposed_array, cmap='gray')
    plt.show()





def make_net1():
    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(16, (4, 4), pad=1, input_size=(1, 128, 128)), initializer=Initializer.He())
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(32, (4, 4), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(128, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dense(512, initializer=Initializer.He(), activation=Layer.Relu()))
    net.add_layer(Layer.Dropout(0.4))
    net.add_layer(Layer.Dense(512, initializer=Initializer.He(), activation=Layer.Relu()))
    net.add_layer(Layer.Dropout(0.4))
    net.add_layer(Layer.Dense(3))
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.SoftmaxWithLoss())
    return net


def make_net3():
    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, input_size=(3, 128, 128)), initializer=Initializer.He())
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.2))
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(32, (3, 4), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Conv2D(128, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dense(512 , initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Dense(1024, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Dense(3))
    net.add_layer(Layer.SoftmaxWithLoss())
    return net



def make_net4():
    net = MultiLayerNet(is_use_dropout=False, weight_decay_lambda=0.1)
    net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, input_size=(3, 128, 128)), initializer=Initializer.He())
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, initializer=Initializer.He()))
    net.add_layer(Layer.BatchNormalization())
    net.add_layer(Layer.Relu())
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dropout(0.5))

def make_net5():
    net = MultiLayerNet(is_use_dropout=False)
    net.load_model("./idol_result/train_weight_2021-04-24 123923.npz")
    return net






def make_net2():

    # net.add_layer(Layer.Conv2D(16, (3, 3), pad=1, input_size=(1, 128, 128)), initializer=Initializer.He())
    # net.add_layer(Layer.BatchNormalization())
    # net.add_layer(Layer.Relu())
    # net.add_layer(Layer.Dropout(0.2))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, initializer=Initializer.He()))
    # net.add_layer(Layer.BatchNormalization())
    # net.add_layer(Layer.Relu())
    # net.add_layer(Layer.Dropout(0.2))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Conv2D(64, (3, 3), pad=1, initializer=Initializer.He()))
    # net.add_layer(Layer.BatchNormalization())
    # net.add_layer(Layer.Relu())
    # net.add_layer(Layer.Dropout(0.2))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Conv2D(128, (3, 3), pad=1, initializer=Initializer.He()))
    # net.add_layer(Layer.BatchNormalization())
    # net.add_layer(Layer.Relu())
    # net.add_layer(Layer.Dropout(0.2))
    # net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    # net.add_layer(Layer.Dense(512, initializer=Initializer.He()))
    # net.add_layer(Layer.BatchNormalization())
    # net.add_layer(Layer.Relu())
    # net.add_layer(Layer.Dropout(0.5))
    net = MultiLayerNet(is_use_dropout=False)
    net.add_layer(Layer.Conv2D(32,(3,3), input_size=(1, 128, 128)), initializer=Initializer.He(), activation=Layer.Relu())
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
    net.add_layer(Layer.Dropout(0.5))
    net.add_layer(Layer.Dense(3))
    net.add_layer(Layer.SoftmaxWithLoss())
    return net


# def make_net2():
#     net = MultiLayerNet(is_use_dropout=False)
#     net.add_layer(Layer.Conv2D(32, (3, 3), pad=1, input_size=(1, 64, 64)), initializer=Initializer.He(),
#                   activation=Layer.Relu())
#     net.add_layer(Layer.Pooling(pool_h=2, pool_w=2, stride=2))
#     net.add_layer(Layer.Dense(128, initializer=Initializer.He(), activation=Layer.Relu()))
#     net.add_layer(Layer.Dense(2, initializer=Initializer.He()))
#     net.add_layer(Layer.SoftmaxWithLoss())
#
#     return net


def shuffle_dataset(x, t):
    from sklearn.utils import shuffle
    x, t = shuffle(x, t)
    return x, t

print("Loading Dataset..")

load_data = np2.load("./idol_images_gray_128_v2.npz", allow_pickle=True)
x_train = load_data["x_train"]

t_train = load_data["t_train"].astype(np.int64)

x_test = load_data["x_test"]
t_test = load_data["t_test"].astype(np.int64)

x_train, t_train = shuffle_dataset(x_train, t_train)
x_test, t_test = shuffle_dataset(x_test, t_test)





x_train = x_train / 255
x_test = x_test / 255


def train_model():
    net = make_net2()

    for k, v in net.params.items():
        print(k, v.shape)
    print(net.layers)
    print(net.hiddenSizeList)

    with Timer(name="jjy 3-Dense Layer"):
        result = net.train(
            x_train, t_train, x_test, t_test, batch_size=32, iters_num=937, print_epoch=1, evaluate_limit=500,
            is_use_progress_bar=True, save_model_each_epoch=1, save_model_path="./idol_result",
            optimizer=Optimizer.Adam(lr=0.0001))

    import pickle
    import datetime
    #
    ## Save pickle
    with open(f"train_data_{str(datetime.datetime.now())[:-7].replace(':', '')}.pickle", "wb") as fw:
        pickle.dump(result, fw)
    net.save_model("weight_idol_f55.npz")

    print("============================================")


train_model()



# load_trained_model()


# %%
import jsonpickle as jsp
from inspect import currentframe
import numpy as np2
import jjy.framework.initializer as Initializer
import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
from tqdm import tqdm
import os
import json
import math


def debug_print(arg):
    return
    frameinfo = currentframe()
    print(frameinfo.f_back.f_lineno, ":", arg)


import datetime
# %%

import sys

sys.path.append("../../")

# %%
import re
from jjy.framework.cuda import xp as np
from jjy.framework.cuda import GPU_ENABLE, GPU_MODE
from collections import OrderedDict


# %%


class MultiLayerNet:

    def __init__(self, weight_decay_lambda=0, is_use_dropout=False, dropout_ratio=0.5):

        self.is_gpu = False

        self.weight_decay_lambda = weight_decay_lambda
        self.is_use_dropout = is_use_dropout
        self.dropout_ratio = dropout_ratio

        self.params = {}
        self.layers = OrderedDict()
        self.lastLayer = Layer.IdentityWithLoss()
        self.hiddenSizeList = []
        self.trainable_param_keys = []
        self.prevDenseLayer = None
        self.prevConvLayer = None

        self.pre_channel_num = None

        self.added_layer_list = []

    def reset(self):
        pass

    def add_layer(self, layer, **kwargs):

        is_direct = True
        if "is_direct" in kwargs and kwargs["is_direct"] == False:
            is_direct = False

        if not isinstance(layer, Layer.LayerType):
            raise BaseException("Layer required")

        if type(self.lastLayer) == type(Layer.SoftmaxWithLoss()):
            raise BaseException("Already last layer set")

        layer_len = len(self.layers)

        if isinstance(layer, Layer.Dense):

            input_size = layer.input_size

            if (input_size is None and len(self.hiddenSizeList) > 0):
                #                 input_size = self.prevDenseLayer.hidden_size
                #                 print(self.hiddenSizeList)
                input_size = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
            elif type(input_size) == tuple:
                input_size = int(np.prod(np.asarray(input_size)))
            else:
                self.hiddenSizeList.append(input_size)

            hidden_size = layer.hidden_size

            weight_init_std = 0.01
            initializer = layer.initializer

            pre_node_nums = input_size

            if isinstance(initializer, Initializer.Std):
                weight_init_std = initializer.std
            elif isinstance(initializer, Initializer.He):
                weight_init_std = np.sqrt(2.0 / pre_node_nums)
            elif isinstance(initializer, Initializer.Xavier):
                weight_init_std = np.sqrt(1.0 / pre_node_nums)

            self.hiddenSizeList.append(hidden_size)
            #             print(input_size)
            #             print("input", self.hiddenSizeList)
            #             print(input_size, hidden_size)
            # print("SDFSFD", input_size, hidden_size)

            # if type(input_size) == tuple:

            self.params[f"W{layer_len}"] = weight_init_std * np.random.randn(input_size, hidden_size)
            self.params[f"b{layer_len}"] = np.zeros(hidden_size)

            self.layers[f"Affine{layer_len}"] = Layer.Affine(self.params[f"W{layer_len}"], self.params[f"b{layer_len}"],
                                                             trainable=layer.trainable)

            if layer.trainable is True:
                self.trainable_param_keys.append(f"W{layer_len}")
                self.trainable_param_keys.append(f"b{layer_len}")

            self.prevDenseLayer = layer

            if layer.activation is not None:
                self.add_layer(layer.activation, is_direct=False)

            if self.is_use_dropout:
                layer_len = len(self.layers)
                self.layers['Dropout' + str(layer_len)] = Layer.Dropout(self.dropout_ratio)


        elif isinstance(layer, Layer.Conv2D):

            if self.pre_channel_num == None:
                self.pre_channel_num = layer.input_size[0]

            input_size = layer.input_size

            if input_size is None and len(self.hiddenSizeList) > 0:
                input_size = self.hiddenSizeList[-1]
                # print(input_size)
                # print(self.hiddenSizeList)

            filter_num = layer.filter_num

            weight_init_std = 0.01
            initializer = layer.initializer

            if self.prevConvLayer is None:
                pre_node_nums = np.prod(np.asarray(layer.filter_size))
            else:
                pre_node_nums = self.prevConvLayer.filter_num * np.prod(np.asarray(layer.filter_size))

            if isinstance(initializer, Initializer.Std):
                weight_init_std = initializer.std
            elif isinstance(initializer, Initializer.He):
                weight_init_std = np.sqrt(2.0 / pre_node_nums)
            elif isinstance(initializer, Initializer.Xavier):
                weight_init_std = np.sqrt(1.0 / pre_node_nums)

            # print(pre_node_nums)

            # print(layer)
            # print(layer.filter_size, layer.pad, layer.stride, input_size, "mu")

            conv_output_size = (
                layer.filter_num, int((input_size[1] - layer.filter_size[0] + 2 * layer.pad) / layer.stride + 1),
                int((input_size[2] - layer.filter_size[1] + 2 * layer.pad) / layer.stride + 1))
            #
            # print(conv_output_size, "!!!!!!")

            self.hiddenSizeList.append(conv_output_size)

            # print("input", self.hiddenSizeList)

            debug_print(input_size)

            self.params[f"W{layer_len}"] = weight_init_std * np.random.randn(layer.filter_num, self.pre_channel_num,
                                                                             layer.filter_size[0], layer.filter_size[1])

            self.pre_channel_num = layer.filter_num

            # print("??")

            self.params[f"b{layer_len}"] = np.zeros(layer.filter_num)
            debug_print(input_size)




            self.layers[f"Convolution{layer_len}"] = Layer.Convolution(self.params[f"W{layer_len}"],
                                                                       self.params[f"b{layer_len}"], layer.stride,
                                                                       layer.pad, trainable=layer.get_trainable())

            if layer.trainable is True:
                self.trainable_param_keys.append(f"W{layer_len}")
                self.trainable_param_keys.append(f"b{layer_len}")

            debug_print(input_size)
            self.prevConvLayer = layer

            debug_print(input_size)

            if layer.activation is not None:
                self.add_layer(layer.activation, is_direct=False)

            if self.is_use_dropout:
                self.layers['Dropout' + str(layer_len)] = Layer.Dropout(self.dropout_ratio)
            debug_print(input_size)

        elif isinstance(layer, Layer.Pooling):
            prevLayer = self.prevConvLayer
            #             conv_output_size = (input_size - filter_size + 2*filter_pad) / filter_stride + 1

            conv_output_size = self.hiddenSizeList[-1]

            #             print(conv_output_size)
            pool_output_size = (
                prevLayer.filter_num, int((conv_output_size[1] / layer.stride)),
                int((conv_output_size[2] / layer.stride)))
            # print("yaya", conv_output_size, pool_output_size)

            self.hiddenSizeList.append(pool_output_size)

            self.layers[f"Pooling{layer_len}"] = layer

        elif isinstance(layer, Layer.Dropout):
            self.layers[f"Dropout{layer_len}"] = Layer.Dropout(layer.dropout_ratio)

        elif isinstance(layer, Layer.Relu):

            self.layers[f"Relu{layer_len}"] = Layer.Relu()

        elif isinstance(layer, Layer.Sigmoid):

            self.layers[f"Sigmoid{layer_len}"] = Layer.Sigmoid()

        elif isinstance(layer, Layer.SoftmaxWithLoss):

            #             self.layers[f"SoftmaxWithLoss{layer_len}"] = Layer.SoftmaxWithLoss()
            self.lastLayer = layer

        elif isinstance(layer, Layer.IdentityWithLoss):

            #             self.layers[f"SoftmaxWithLoss{layer_len}"] = Layer.SoftmaxWithLoss()
            self.lastLayer = layer


        elif isinstance(layer, Layer.BatchNormalization):
            # print(self.hiddenSizeList)
            # print(np.prod(np.asarray([self.hiddenSizeList[-1]])))
            # if type(self.hiddenSizeList[-1]) == int:
            #     n = self.hiddenSizeList[-1]
            # else:
            #     n = self.prevConvLayer.filter_num

            n = int(np.prod(np.asarray(self.hiddenSizeList[-1])))
            # print(n)
            layer.__init__(gamma=np.ones(n),
                           beta=np.zeros(n),
                           running_mean=layer.running_mean,
                           running_var=layer.running_var)
            self.layers[f"BatchNormal{layer_len}"] = layer
            # self.layers[f"BatchNormal{layer_len}"] = Layer.BatchNormalization(
            #     gamma=np.ones(n),
            #     beta=np.zeros(n),
            #     running_mean=layer.running_mean,
            #     running_var=layer.running_var
            # )
        # print(layer, self.hiddenSizeList)
        elif isinstance(layer, Layer.FreezePrev):
            for key, layer in self.layers.items():
                if isinstance(layer, Layer.BatchNormalization) or isinstance(layer, Layer.Affine) or isinstance(layer,
                                                                                                                Layer.Convolution):
                    self.layers[key].trainable = False
                    print(self.layers[key].trainable)

            self.trainable_param_keys = []
            return

        if is_direct:
            self.added_layer_list.append(layer)

    def predict(self, x, train_flg=False):
        for layer in self.layers.values():

            if isinstance(layer, Layer.BatchNormalization) or isinstance(layer, Layer.Dropout):

                x = layer.forward(x, train_flg)
            else:
                # print(layer)
                x = layer.forward(x)

        return x

    # x : 입력 데이터, t : 정답 레이블
    def loss(self, x, t, train_flg=False):
        y = self.predict(x, train_flg)

        weight_decay = 0
        for idx, (name, layer) in enumerate(self.layers.items()):
            if isinstance(layer, Layer.Affine):
                layer_num = name[6:]
                W = self.params['W' + str(layer_num)]
                weight_decay += 0.5 * self.weight_decay_lambda * np.sum(W ** 2)

        return self.lastLayer.forward(y, t) + weight_decay

    def accuracy(self, x, t, batch_size=100):

        correct_cnt = 0
        data_num = int(x.shape[0])

        for i in range(math.ceil(data_num / batch_size)):
            # print("batch", i)
            xx = x[batch_size * i: min(batch_size * (i + 1), data_num)]
            tt = t[batch_size * i: min(batch_size * (i + 1), data_num)]

            yy = self.predict(xx, train_flg=False)
            yy = np.argmax(yy, axis=1)
            if tt.ndim != 1: tt = np.argmax(tt, axis=1)

            correct_cnt += np.sum(yy == tt)

        accuracy = correct_cnt / float(x.shape[0])
        return accuracy

    def gradient(self, x, t, train_flg = True):
        # forward
        self.loss(x, t, train_flg=True)

        # backward
        dout = 1
        dout = self.lastLayer.backward(dout)

        layers = list(self.layers.values())
        layers.reverse()
        for layer in layers:
            if layer.trainable is True or train_flg is False:
                dout = layer.backward(dout)

        # 결과 저장
        grads = {}

        for idx, (name, layer) in enumerate(self.layers.items()):
            if isinstance(layer, Layer.Affine):
                layer_num = name[6:]
                grads[f"W{layer_num}"] = layer.dW + self.weight_decay_lambda * self.params['W' + str(layer_num)]
                grads[f"b{layer_num}"] = layer.db
            elif isinstance(layer, Layer.Convolution):
                layer_num = name[11:]
                grads[f"W{layer_num}"] = layer.dW + self.weight_decay_lambda * self.params['W' + str(layer_num)]
                grads[f"b{layer_num}"] = layer.db

        return grads

    def train(self, x_train, t_train, x_test, t_test, **kwargs):  # t_data 는 원 핫 인코딩

        if GPU_ENABLE:
            if GPU_MODE:
                print("GPU로 동작합니다. CPU로 동작하게 하려면 'JJY_GPU_ENABLE' 환경변수의 값을 'false'로 설정해주세요")
            else:
                print("CPU로 동작합니다. GPU로 동작하게 하려면 cupy 라이브러리를 설치해주세요.")
        else:
            print("CPU로 동작합니다. GPU로 동작하게 하려면 'JJY_GPU_ENABLE' 환경변수의 값을 'true'로 설정해주세요")

        optimizer = Optimizer.SGD(lr=0.01)
        if "optimizer" in kwargs:
            if isinstance(kwargs["optimizer"], Optimizer.OptimizerType):
                optimizer = kwargs["optimizer"]
            else:
                raise "err"

        optimizer.__init__(trainable_param_keys=self.trainable_param_keys)

        iters_num = kwargs.get("iters_num", None)
        if iters_num is None:
            return ValueError("iters_num must be set")
        elif not (iters_num > 0):
            raise ValueError("iters_num must be > 0 ")

        batch_size = kwargs.get("batch_size", None)
        if batch_size is None:
            return ValueError("batch_size must be set")
        elif not (batch_size > 0):
            raise ValueError("batch_size must be > 0")

        print_epoch = kwargs.get("print_epoch", 1)
        if print_epoch is None:
            print_epoch = 9999999999999999999
        elif not (print_epoch > 0):
            raise ValueError("print_epoch must be > 0")

        save_model_each_epoch = kwargs.get("save_model_each_epoch", None)
        if save_model_each_epoch is None:
            save_model_each_epoch = 9999999999999999999
        elif not (save_model_each_epoch > 0):
            raise ValueError("save_model_each_epoch must be > 0")

        evaluate_limit = kwargs.get("evaluate_limit", None)
        save_model_path = kwargs.get("save_model_path", "")

        is_use_progress_bar = kwargs.get("is_use_progress_bar", False)

        e_x_train, e_x_test, e_t_train, e_t_test = x_train, x_test, t_train, t_test

        if evaluate_limit is not None:
            e_x_train, e_x_test, e_t_train, e_t_test = np.asarray(x_train[:evaluate_limit]), np.asarray(
                x_test[:evaluate_limit]), np.asarray(t_train[
                                                     :evaluate_limit]), np.asarray(t_test[
                                                                                   :evaluate_limit])

        output_type = "regression"
        if type(self.lastLayer) == type(Layer.SoftmaxWithLoss()):
            output_type = "class"

        train_size = x_train.shape[0]

        train_loss_list = []
        train_acc_list = []
        test_acc_list = []

        iter_per_epoch = max(train_size / batch_size, 1)
        print("iter_per_epoch :", iter_per_epoch)

        print(f"repeat {int(iters_num // (max(train_size / batch_size, 1)))} epoch")

        if output_type == "regression":
            print("epoch | loss")
        else:
            print("epoch | train_acc | test_acc | loss")

        cnt = 1

        if is_use_progress_bar:
            iterator = tqdm(range(iters_num), mininterval=1)
        else:
            iterator = range(iters_num)

        for i in iterator:
            # with nostdout():
            # with redirect_to_tqdm():

            # print(i)

            batch_mask = np2.random.choice(train_size, batch_size)
            x_batch = np.asarray(x_train[batch_mask])
            t_batch = np.asarray(t_train[batch_mask])

            grad = self.gradient(x_batch, t_batch, train_flg = False)

            # 갱신

            optimizer.update(self.params, grad)

            loss = self.loss(x_batch, t_batch)
            train_loss_list.append(loss)

            if int(cnt * iter_per_epoch) == i:

                if save_model_each_epoch is not None and cnt % save_model_each_epoch == 0:
                    self.save_model(
                        save_model_path + f"/train_weight_{str(datetime.datetime.now())[:-7].replace(':', '')}.npz")

                if cnt % print_epoch == 0:

                    train_acc = self.accuracy(e_x_train, e_t_train, batch_size=batch_size)
                    test_acc = self.accuracy(e_x_test, e_t_test, batch_size=batch_size)
                    train_acc_list.append(train_acc)
                    test_acc_list.append(test_acc)
                    if output_type == "regression":
                        print(f"epoch {cnt}:", train_loss_list[-1])
                    else:
                        # log.info(f"epoch {cnt}:")
                        print(f"epoch {cnt}:", format(float(train_acc), ".4f"), " | ", format(float(test_acc), ".4f"),
                              " | ", format(float(loss), ".4f"),
                              flush=True)

                cnt += 1

        print("===================")
        train_acc = self.accuracy(e_x_train, e_t_train, batch_size=batch_size)
        test_acc = self.accuracy(e_x_test, e_t_test, batch_size=batch_size)
        if output_type == "regression":
            print(f"epoch {cnt - 1}:", train_loss_list[-1])
        else:
            print(f"epoch {cnt - 1}:", format(float(train_acc), ".4f"), " | ", format(float(test_acc), ".4f"))

        return {
            "train_loss_list": train_loss_list,
            "train_acc_list": train_acc_list,
            "test_acc_list": test_acc_list
        }

    def save_model(self, path=None):
        save_data = OrderedDict()
        params = self.params

        save_data["NetConfig"] = {
            "weight_decay_lambda": self.weight_decay_lambda,
            "is_use_dropout": self.is_use_dropout,
            "dropout_ratio": self.dropout_ratio
        }

        other_params = {}

        layer_info = []

        for layer in self.added_layer_list:
            t = layer
            if isinstance(t, Layer.SoftmaxWithLoss):
                t = Layer.SoftmaxWithLoss()
            elif isinstance(t, Layer.IdentityWithLoss):
                t = Layer.IdentityWithLoss()
            elif isinstance(t, Layer.BatchNormalization):
                t = Layer.BatchNormalization()


            elif isinstance(t, Layer.Pooling):

                t = Layer.Pooling(t.pool_h, t.pool_w, t.stride, t.pad)

            s = jsp.encode(t)
            layer_info.append(s)

        for layer_name, layer in self.layers.items():
            if isinstance(layer, Layer.BatchNormalization):
                layer_idx = int(re.findall(r'[0-9]+', layer_name)[-1])
                other_params[f"BN_m{layer_idx}"] = layer.running_mean
                other_params[f"BN_v{layer_idx}"] = layer.running_var

        save_data["LayerInfo"] = layer_info
        # for layer_idx, layer in enumerate(self.layers.values()):
        #     if isinstance(layer, Layer.BatchNormalization):
        #         params[f"BN_m{layer_idx}"] = layer.running_mean
        #         params[f"BN_v{layer_idx}"] = layer.running_var

        save_data["Params"] = params
        save_data["OtherParams"] = other_params

        if path is None:
            path = f"train_weight_{str(datetime.datetime.now())[:-7].replace(':', '')}.npz"
        # print(save_data)

        np.savez_compressed(path, **save_data)
        print(f"Weight was saved at {path}")

    def load_model(self, path):
        import numpy as np2
        print(f"Loading Model... {path}")
        load_data = np2.load(path, allow_pickle=True)
        other_config_data = dict(load_data.get("NetConfig").item())
        self.__init__(**other_config_data)
        layer_info = load_data.get("LayerInfo")
        for layer_encoded in layer_info:
            layer = jsp.decode(layer_encoded)
            if hasattr(layer, "trainable"):
                pass
            else:
                layer.trainable = True


            # print(layer_encoded)

            self.add_layer(layer)

        self.params = load_data.get("Params").item()
        other_params = load_data.get("OtherParams").item()

        for param_name, param_value in self.params.items():
            # print(param_name)

            layer_idx = int(re.findall(r'[0-9]+', param_name)[0])
            param_kind = re.sub(r'[0-9]+', '', param_name)
            target_layer = list(self.layers.items())[layer_idx][1]
            if param_kind == "W":
                target_layer.W = param_value
            elif param_kind == "b":
                target_layer.b = param_value
            elif param_kind == "BN_m":
                target_layer.running_mean = param_value
            elif param_kind == "BN_v":
                target_layer.running_var = param_value

        for param_name, param_value in other_params.items():
            layer_idx = int(re.findall(r'[0-9]+', param_name)[0])
            param_kind = re.sub(r'[0-9]+', '', param_name)
            target_layer = list(self.layers.items())[layer_idx][1]
            if param_kind == "BN_m":
                target_layer.running_mean = param_value

                # print(target_layer)
                # print(target_layer.running_mean)
            elif param_kind == "BN_v":
                target_layer.running_var = param_value

# %%


# %%



# %%

import sys

import jjy.framework.layer as Layer
import jjy.framework.optimizer as Optimizer
from jjy.framework.functions import *
from jjy.framework.network import MultiLayerNet
import jjy.framework.initializer as Initializer
from jjy.dataset.mnist import load_mnist

import cupy as np


def make_sample_data_set():
    """
        x는 0~999999 정수이다.
        정답은 x를 100000으로 나눴을 때의 몫이다.
    """

    x = np.random.randint(999999, size=(10000, 1))

    t_data = x.flatten() // 100000

    # t_data 원핫 인코딩 코드
    num = np.unique(t_data, axis=0)
    num = num.shape[0]
    t = np.eye(num)[t_data]

    return x, t


def make_sample_data_set_regression():
    x = np.random.randint(9, size=(300, 1))
    y = np.dot(x, np.array([3]))
    t = np.reshape(y, (y.shape[0], 1))

    return x, t


def make_sample_data_set_regression2():
    x = np.random.randint(999999, size=(300, 2))
    y = np.dot(x, np.array([3, -2])) + 7
    #     y =  y + (4 * np.random.random_sample((1,y.shape[0])) - 2).flatten()
    t = np.reshape(y, (y.shape[0], 1))

    return x, t


(x_train, t_train), (x_test, t_test) = load_mnist(normalize=True, one_hot_label=True)

print(x_train.shape, t_train.shape, x_test.shape, t_test.shape)

x_data = np.append(x_train, x_test, axis=0)
t_data = np.append(t_train, t_test, axis=0)

net = MultiLayerNet(is_use_dropout=True, dropout_ratio=0.2)
net.add_layer(Layer.Dense(30, input_size=784, initializer=Initializer.He()))
net.add_layer(Layer.BatchNormalization())
net.add_layer(Layer.Relu())
net.add_layer(Layer.Dense(64), initializer=Initializer.He())
net.add_layer(Layer.BatchNormalization())
net.add_layer(Layer.Relu())
net.add_layer(Layer.Dense(64), initializer=Initializer.He())
net.add_layer(Layer.BatchNormalization())
net.add_layer(Layer.Relu())
net.add_layer(Layer.Dense(64), initializer=Initializer.He())
net.add_layer(Layer.BatchNormalization())
net.add_layer(Layer.Relu())
net.add_layer(Layer.Dense(64), initializer=Initializer.He())
net.add_layer(Layer.BatchNormalization())
net.add_layer(Layer.Relu())
net.add_layer(Layer.Dense(10, initializer=Initializer.He()
                          , activation=Layer.SoftmaxWithLoss()))

result = net.train(
    x_train, t_train, x_test, t_test, batch_size=300, iters_num=1000, print_epoch=1, is_use_progress_bar=True,
    save_model_each_epoch=1, save_model_path="./dog_cat_result",
    optimizer=Optimizer.Adam(lr=0.01))

print("done!")

print(net.params)
# print(net.layers)



# net.save_model()

# net.load_model("weight.npz")

print("============================================")

# net2 = MultiLayerNet()
# net2.load_model("weight.npz")
#
# # print(net.layers)
# # print(net.layers["BatchNormal2"].running_mean)
#
# print(net2.accuracy(x_test, t_test))
